/*
   What does the serial tab do?, magic.
   But in all seriousness, Serial command line allows control of the printhead over either the built in USB to serial or external serial
   It gives all commands required to print with the printhead at the highest speed possible over serial.

   basic structure, 3 letter code (in caps), space, then a value in either base 64 or base 10. '\n', '\r' or '&' ends the line
   If you want every typed line to function as a single command, make sure you have new line or carriage return enabled
   Alternatively you can end each line yourself by using '&' after each command, the serial function is not that picky

   Base 10 (B10) is simply 0-9. Depending on the function the size of the variable varies
     Base 10 takes up to 9 digits. Any more will be cut off from the most significant side. This may cause errors
   Base 64 (B64) is A-Z, a-z, 0-9 and + and / (0-63, or 000000 to 111111).
   TXT (T) is Ascii characters, at most around 70
   This is to increase the number of bits per character from 4 (0-7) to 6

   example line for B10: "PMD 1" set printmode to 1
   example line for B64: "PVR AI8UHT6YJJY89abyth9+enigf74/hytJLS9+qy65GLP096gGrh" Print whatever bits are in there

   [10]SRT: Start Serial mode, creates the buffer for serial (B10 0: for just response or 1: for create all values)

   [510]GTP: Get temperature (B10: mode, 1 for celcius, 101 for Fahrenheit (also 2)) (If address is not in 0, temperature will error)
   [520]NTA: Nozzle table address (B10: nozzle to get)
   [530]NTP: Nozzle table primitive (B10: nozzle to get)


   [101]PVR: Print burst from value raw (takes 50/25 characters in base 64) (B10 aux value)
   [111]PVT: Print burst from toggle row (takes a variable number of characters in base 64, the sum totaling 300/150) (B10 aux value)
   [121]PAP: Print Address/Primitive mode (Takes 66 base 64 characters (22 blocks of 3) (B10 aux mode)
   [200]PMD: Print mode (pos mode, trigger mode, time mode and many more (see below))
   [210]DPI: (Set) printhead DPI (100, 150, 200, 300, 600)
   [220]SOE: Set odd/even. (B10: 0 for full head, 1 for odd, 2 for even)
   [x230]DST: (Set) printing density. The spacing of the ink in the sweep direction compared to 600DPI (B10 percentage) Only works for pos mode
   [300]HEN: Head enable (B10: 1 for on, 0 for off
   [310]TST: Test printhead
   [320]PHT: Preheat the printhead (*B10: Number of cycles to preheat (defaults to 18k)

   [x]BFN: (add) Buffer next
   [x]BFP: (add) Buffer Pos
   [x]BNX: Buffer next
   [x]BRL: Buffer repeats left
   [x]BCR: Buffer clear
   [x]PBF: Print current buffer line


   [x]SCP: set current position (B10: microns)
   [x]EPI: Encode pulses per inch (B10)

   --low level commands
   [1000]ANX: Address next (B10 1 or 0 for on or off, 10 for on/off)
   [1020]ARS: Address reset (B10 1 or 0 for on or off, 10 for on/off)
   [x1040]PCK: Primitive clock (B10 1 or 0)
   [x1060]PCR: Primitive clear (B10 1 or 0) (active low)
   [x1080]PLS: Pulse primitive (B10 1)
   [x1100]SVL: Shift register value (B10 16bit number to send to shift register)

   //explain odd and even mode
   SOE: The HP45 has 2 rows of nozzles, around 4mm apart. Synchronizing these rows might be too bothersome for some systems
        SOE sets which one or both rows are used. 0 sets both rows to active. 1 sets odd to active, 2 sets even to active
        IMPORTANT: in odd or even mode, the amount of nozzles go to 150, not 300. PAP mode is not affected by SOE

   //explain PVR, PVT and PAP mode
   PVR: Print Value Raw sends which nozzle to activate as raw data. This is most efficient with varying data (like greyscale images) but is always the same size
        The data is sent using Base 64, where A is 000000, B is 000001.... the LSB is the lowest nozzle. A 1 is the nozzle is turned on, a 0 is the nozzle turned off
        If no full data is sent, the missing data is presumed to be 0
   PVT: Print Value Toggle sends the data in a series of numbers. The first number defines how many nozzles need to be turned on, the next how many are turned off, and so on
        This continues until the sum is 300 nozzles. If you need the first nozzle(s) off, send an A (0) first. This also applies to sets of nozzles bigger than 64.
        Setting all 300 nozzles is done as follows: PVT /A/A/A/As (I might have messed up the last character, just set it to / if you need to be safe)
        If the sum is not 300, any remainder will be set to 0, regardless of the last toggle being high or low. Too much data will be ignored
   PAP: Print Address Primitive sends the data with raw address/primitive data. This should only really be used if the controller cannot handle the data
        Paradoxically, PAP mode needs more data sent. Each address is encoded as 3 base 64 characters, for a total of 22 addresses (66 in total)
        If fewer than 66 characters are sent, the remainder will be set to 0
        To get the nozzles for each address and primitive, use the NTA and NTP commands
   B10 aux mode: All of these commands need an additional value besides the inkjet information. They need either a repeat or coordinate in microns
        in B10. Depending on the mode. Trigger mode and timed mode need repeats, how often this line is reapeated before the next one should be fetched
        Pos mode takes a coordinate in microns, where this line should be activated. Coordinates need to be sequential based on the printing direction

   PMD: Printmode signifies how to print with any buffered data. This includes many fancy modes like:
        Comming soon (the best of all modes)
*/

/*
   Todo list
   Add text mode to the interface (text should be buffered a layer higher as raw text as well)
   Add single trigger control to the serial line
   Set trigger pin function
   Set encoder pins function
   Add position control and timing modes to the functions (needed for buffered functions)
   Add buffer mode to printing commands
   Buffer Add
   Buffer clear
   Buffer left
   Buffer force next
   Buffer remaining (forced callback on certain points)
   Add text mode to printing modes (add separate font library)
*/

//#define serial_baud 2500000 //Serial over serial port 1 baudrate
#define serial_buffer_size 90 //the size of the character buffer, always needs to be able to contain 1 full line of code, but not too big
#define serial_max_b10_size 9 //the max size of a base 10 input (b10 is stored in a single int32
#define serial_max_b64_size 70 //the max size of a base 64 input (b64 is stored in an uint8_t[#] array

//table that defines what reading modes are called in what order. Goes up to 4, ends in -1 (can easily be increased in size)
const int8_t serial_change_table[10][5] = {
  {10, -1, -1, -1, -1}, //0 B10
  {60, -1, -1, -1, -1}, //1 B64
  {60, 10, -1, -1, -1}, //2 B64, B10
  {60, 10, 10, -1, -1}, //3 B64, B10, B10
  { -1, -1, -1, -1, -1}, //4
  { -1, -1, -1, -1, -1}, //5
  {30, -1, -1, -1, -1}, //6 Text
  {10, 30, -1, -1, -1}, //7 B10, Text //Text has few available end characters, so it should be last for stability
  { -1, -1, -1, -1, -1}, //8
  { -1, -1, -1, -1, -1}, //9
};

//variables
uint8_t serial_mode = 0; //0 is startup mode, 1 is full serial mode.
char serial_buffer[serial_buffer_size]; //buffer for storing serial data
int16_t serial_buffer_read_end = 0;
char serial_temp_buffer[serial_buffer_size]; //buffer for storing serial data

int32_t serial_base10_value; //the variable for storing the base 10 value
uint8_t serial_base64_value[serial_max_b64_size]; //the variable for storing the base 64 value (6 LSB's are filled)
uint16_t serial_base64_size = 0; //how many characters there are in the base 64 value

uint8_t serial_change_pos; //where in the change table the user is


//list of op codes with their values
//general rule, more important or faster codes should be earlier in the list
//values ending in 0 take base 10 input, values ending in 1 take base 64 values (more left for future input types)
//values ending in 2 take a base64 first, then a base10,
#define serial_op_codes 19 //the number of op codes in the list
const uint16_t serial_op_code_value[serial_op_codes] = {
  10, //0: SRT: start
  101, //PVR: print value raw
  111, //PVT: print value toggle
  121, //PAP: print address primitive
  200, //PMD: printmode
  210, //DPI: DPI
  220, //SOE: Set odd even
  300, //HEN: Head enable
  310, //TST: Test printhead
  320, //PHT: Preheat printhead
  510, //GTP: Get temperature
  520, //NTA: Nozzle table address
  530, //NTP: Nozzle table primitive
  1000, //ANX: Address next
  1020, //ARS: Address Reset
  1040, //PCK: Primitive clock
  1060, //PCR: Primitive clear
  1080, //PLS: Pulse primitive
  1100, //SVL: Shiftregister value
};

const char serial_op_code_name[serial_op_codes][3] = {
  {'S', 'R', 'T'}, //SRT: start
  {'P', 'V', 'R'}, //PVR: print value raw
  {'P', 'V', 'T'}, //PVT: print value toggle
  {'P', 'A', 'P'}, //PAP: print address primitive
  {'P', 'M', 'D'}, //PMD: printmode
  {'D', 'P', 'I'}, //DPI: head dpi
  {'S', 'O', 'E'}, //SOE: set odd even
  {'H', 'E', 'N'}, //HEN: head enable
  {'T', 'S', 'T'}, //TST: test printhead
  {'P', 'H', 'T'}, //PHT: preheat
  {'G', 'T', 'P'}, //GTP: Get temperature
  {'N', 'T', 'A'}, //NTA: Nozzle table address
  {'N', 'T', 'P'}, //NTP: Nozzle table primitive
  {'A', 'N', 'X'}, //ANX: Address Next
  {'A', 'R', 'S'}, //ARS: Address Reset
  {'P', 'C', 'K'}, //PCK: Primitive clock
  {'P', 'C', 'R'}, //PCR: Primitive clear
  {'P', 'L', 'S'}, //PLS: Primimitive pulse
  {'S', 'V', 'L'}, //SVL: Shiftregister value
};





//Serial update reads from serial and performs any task asked in the serial
//return codes:
//1: no more end character found. Function always should end on 1

//errors
//-10: buffer full while calling this function

int8_t SerialUpdate() {
  int32_t temp_serial_available = Serial.available();

  //fill buffer if possible
  if (temp_serial_available) {
    //Serial.print("Available: "); Serial.println(temp_serial_available);
    //get data
    uint16_t temp_data_to_read; //figure out how much still fits in the buffer
    if (serial_buffer_read_end >= serial_buffer_size - 1) { //check to see if there is still space in the buffer
      //Serial.print("buffer full");
      return -10;
    }
    if (temp_serial_available < serial_buffer_size - serial_buffer_read_end) { //if there is less in memory than fits in the buffer, write all
      temp_data_to_read = temp_serial_available;
    }
    else { //if there is more in memory than fits the buffer, write until full
      temp_data_to_read = serial_buffer_size - serial_buffer_read_end;
    }
    //Serial.print("data to read: "); Serial.println(temp_data_to_read);

    Serial.readBytes(serial_temp_buffer, temp_data_to_read); //write this much data to the temp buffer
    uint16_t temp_read_pos = 0; //make a counter to read from the temp_buffer
    while (1) { //move the data from the temp buffer to the final buffer
      serial_buffer[serial_buffer_read_end] = serial_temp_buffer[temp_read_pos]; //write from temp buffer to serial buffer
      serial_buffer_read_end++; //go to the next position in the serial buffer
      temp_read_pos ++;  //increment the temp read buffer
      temp_data_to_read--; //decrement the number of chars left to read
      if (temp_data_to_read <= 0) break; //if there is no data left, break from while
    }

    //Serial.println("Response:"); //debug data
    //for (uint16_t t = 0; t < serial_buffer_size; t++){
    //  Serial.print(serial_buffer[t]);
    //}
    //Serial.println("\nEnd of response");

    //scan for ends of a line
    while (1) {
      int16_t temp_end_of_line = -1;
      for (uint16_t e = 0; e < serial_buffer_read_end; e++) {
        //Serial.print(serial_buffer[e]);
        if (SerialLineEndCharacter(serial_buffer[e])) { //if line end is found, note position and start reading
          temp_end_of_line = e;
          //Serial.print("end of line found: "); Serial.println(temp_end_of_line);
          break;
        }
      }
      if (temp_end_of_line == -1) { //if the end of line was not found (-1 is still -1)
        if (serial_buffer_read_end < serial_buffer_size - 1) { //if no line end is found, and there is space in the buffer left, break
          return 0;
        }
        else { //if no line end is found, and the buffer is full, throw an error, crash, burn, end civilization (or perhaps fix it)
          return -10;
          //Also, panic
        }
      }

      int16_t temp_op_code = -1; //line op code
      int16_t temp_op_start, temp_op_end;
      int16_t temp_serial_read_state = 0;
      int8_t temp_negative_flag = 1; //negative value, defaults to positive
      for (uint16_t r = 0; r < temp_end_of_line; r++) { //start decoding the line (end of line char itself is ignored
        //op code block -------------------------------------------------
        if (temp_serial_read_state == 0) { //0: pre header state (spaces are ignored)
          if (SerialIsCharacter(serial_buffer[r])) { //if a readable Op code character is found, go to read op code mode.
            temp_serial_read_state = 1; //set to read a command
            serial_change_pos = 0; //set change table pos to 0
            temp_op_start = r; //set start of op code
          }
        }
        if (temp_serial_read_state == 1) { //1: header state (read until terminating character)
          if (serial_buffer[r] == ' ') { //if space, end of value.
            temp_op_end = r; //set end of op code
            if (temp_op_end - temp_op_start == 3) { //check if length of op code is correct (3 characters)
              char temp_op_code_array[3];
              uint8_t temp_val = 0; //create a variable to write into the small op code array
              for (uint8_t c = temp_op_start; c < temp_op_end; c++) { //write all chars to an array
                temp_op_code_array[temp_val] = serial_buffer[c];
                temp_val++; //advance to the next value in the array
              }
              temp_op_code = SerialGetOpCode(temp_op_code_array); //get op code value from op code
              //Serial.print("Op code: "); Serial.println(temp_op_code);
              int32_t temp_code = temp_op_code % 10;
              //set new read mode based on change table
              temp_serial_read_state = serial_change_table[temp_code][serial_change_pos];
              serial_change_pos ++;
            }
            else temp_serial_read_state = -1; //error state, no more functions
          }
        }
        //end of op code block -------------------------------------------------

        //b10 block -------------------------------------------------
        if (temp_serial_read_state == 10) {  //10: pre b10 state (spaces are ignored)
          if (SerialIsB10(serial_buffer[r])) { //if a readable Op code character is found, go to read op code mode.
            temp_negative_flag = 1; //set negative flag to 1
            if (serial_buffer[r] == '-') { // if the first character is '-', ignore start for one more character
              temp_negative_flag = -1; //set negative flag to -1
            }
            else { //if first character is not -
              temp_serial_read_state = 11; //set to read a command
              temp_op_start = r; //set start of op code
              serial_base10_value = 0; //set base 10 value to 0
            }
          }
          else if (serial_buffer[r] != ' ') temp_serial_read_state = -1; //if the character leading to the B10 is not a space
        }

        if (temp_serial_read_state == 11) { //find end character of b10 value
          //look for terminating conditions
          if (SerialIsB10(serial_buffer[r]) == 0 || serial_buffer[r] == '-') { //if an end of block character is found (no match with b10) (12)
            temp_op_end = r - 1;
            temp_serial_read_state = 12;
          }
          if (r == temp_end_of_line - 1) { //if the end of the line is reached (13).
            temp_op_end = r;
            temp_serial_read_state = 13;
          }
        }

        if (temp_serial_read_state == 12 || temp_serial_read_state == 13) { //12,13 create the final b10 value and go to next mode
          //create b10 value
          if (temp_op_end - temp_op_start >= serial_max_b10_size) { //check if the size is not bigger than b10 can handle
            //Serial.println("Overflow detected");
            temp_op_start = temp_op_end - serial_max_b10_size;
            temp_op_start ++;
          }
          uint32_t temp_base10_power = 1; //the power value of the base10 value
          uint32_t temp_calc;
          for (int8_t c = temp_op_end; c >= temp_op_start; c--) { //loop through all values found
            temp_calc = SerialB10Lookup(serial_buffer[c]); //read value from array
            //Serial.print("value added: "); Serial.println(temp_calc);
            temp_calc *= temp_base10_power; //add the power to it
            serial_base10_value += temp_calc; //add this to the base 10 value
            temp_base10_power *= 10; //add a power
          }
          //Serial.print("value: "); Serial.println(serial_base10_value);
          if (temp_serial_read_state == 13) { //end of line is reached, go out of the for loopt
            temp_serial_read_state = -1;
            break;
          }
          int32_t temp_code = temp_op_code % 10;
          //set new read mode based on change table
          temp_serial_read_state = serial_change_table[temp_code][serial_change_pos];
          serial_change_pos ++;

        }
        //end of b10 block -------------------------------------------------

        //B64 read block -------------------------------------------------
        if (temp_serial_read_state == 60) {  //60: pre b64 state (spaces are ignored)
          if (SerialIsB64(serial_buffer[r])) { //if a readable Op code character is found, go to read op code mode.
            temp_serial_read_state = 61; //set to read a command
            serial_base64_size = 0; //reset b64 size
            temp_op_start = r; //set start of op code
            for (uint16_t b = 0; b < serial_max_b64_size; b++) { //reset base 64 value
              serial_base64_value[b] = 0;
            }
          }
          else if (serial_buffer[r] != ' ') temp_serial_read_state = -1; //if the character leading to the B64 is not a space
        }

        if (temp_serial_read_state == 61) { //61: b64 read end state (read until terminating character)
          //look for terminating conditions
          if (SerialIsB64(serial_buffer[r]) == 0) { //if an end of block character is found (no match with b64) (62)
            temp_op_end = r - 1;
            temp_serial_read_state = 62;
          }
          if (r == temp_end_of_line - 1) { //if the end of the line is reached (63).
            temp_op_end = r;
            temp_serial_read_state = 63;
          }
        }

        if (temp_serial_read_state == 62 || temp_serial_read_state == 63) { //62, 63: b64 finishing state
          //create b10 value
          if (temp_op_end - temp_op_start >= serial_max_b64_size) { //check if the size is not bigger than b10 can handle
            //Serial.println("Overflow detected");
            temp_op_start = temp_op_end - serial_max_b64_size;
            temp_op_start ++;
          }
          uint32_t temp_base64_pos = 0; //the power value of the base10 value

          //Loop function has Fifo and Filo available. In Fifo, ABCDE will put 0,1,2,3,4 in the array
          //In Lifo, ABCDE will put 4,3,2,1,0 in the array.
          //I am leaning towards Fifo, but both are available should any need arise later on
          //for (int8_t c = temp_op_end; c >= temp_op_start; c--) { //loop through all values found (LIFO)
          for (int8_t c = temp_op_start; c <= temp_op_end; c++) { //loop through all values found (FIFO)
            serial_base64_value[temp_base64_pos] = SerialB64Lookup(serial_buffer[c]); //read value from array
            //Serial.print("value added: "); Serial.println(serial_base64_value[temp_base64_pos]);
            temp_base64_pos++; //advance to next array pos
            serial_base64_size++;
          }
          //Serial.print("value: "); Serial.println(serial_base10_value);
          if (temp_serial_read_state == 63) { //end of line is reached, go out of the for loopt
            temp_serial_read_state = -1;
            break;
          }
          int32_t temp_code = temp_op_code % 10;
          //set new read mode based on change table
          temp_serial_read_state = serial_change_table[temp_code][serial_change_pos];
          serial_change_pos ++;
        }
        //end of b64 block -------------------------------------------------

        //-1, end function
        if (temp_serial_read_state == -1) {
          break;
        }
      }


      //execute command -----------------------------------------------------
      //Serial.println(temp_op_code);
      switch (temp_op_code) {
        case 10: { //start serial mode
            //Serial.print("Starting full printing mode using code: "); Serial.println(serial_base10_value);
          } break;
        case 101: { //print value raw burst
            PrintingDecodeRaw(serial_base64_value, serial_base64_size);
            PrintingPrintArray();
          } break;
        case 111: { //print value toggle burst
            PrintheadEnable();

            int16_t temp_row_start; //row end is r
            uint16_t print_array[22]; //printing array
            int8_t temp_row_width = 50;

            //reset printing array
            for (uint8_t c = 0; c < 22; c++) {
              print_array[c] = 0;
            }

            for (int16_t r = 0; r < 400; r += 2) {
              temp_row_start = r - temp_row_width; //calculate starting point

              for (int16_t n = 0; n < 300; n += 2) { //loop through all nozzles to check their state
                //fill the array with the data if it is within the row to print
                if (n >= temp_row_start && n <= r) {
                  bitWrite(print_array[PrintheadGetAddress(n)], PrintheadGetPrimitive(n), 1);
                }
                else {
                  bitWrite(print_array[PrintheadGetAddress(n)], PrintheadGetPrimitive(n), 0);
                }
              }

              PrintheadInkjetBurstW(print_array); //print the array

              delay(5); //delay to slow the function down
            }

            PrintheadDisable();
          } break;
        case 121: { //print address/primitive burst

          } break;
        case 200: { //set printmode

          } break;
        case 210: { //set printhead DPI
            PrintingSetDPI(serial_base10_value);
          } break;
        case 220: { //set odd even
            PrintingSetOddEven(serial_base10_value);
          } break;
        case 300: { //head enable
            PrintheadSetEnable(serial_base10_value);
          } break;
        case 310: { //test printhead
            SerialFunctionTestPrinthead(serial_base10_value);
          } break;
        case 320: { //preheat printhead
            SerialFunctionGetTemperature(1);
            PrintheadPreheat(serial_base10_value);
            SerialFunctionGetTemperature(1);
          } break;
        case 510: { //get temperature
            SerialFunctionGetTemperature(serial_base10_value);
          } break;
        case 520: { //nozzle table address
            SerialFunctionGetAddressTable(serial_base10_value);
          } break;
        case 530: { //nozzle table primitive
            SerialFunctionGetPrimitiveTable(serial_base10_value);
          } break;
        case 1000: { //Address next
            SerialFunctionAddressNext(serial_base10_value);
          } break;
        case 1020: { //Address Reset
            SerialFunctionAddressReset(serial_base10_value);
          } break;
        case 1040: { //Primitive clock
            SerialFunctionPrimitiveClock(serial_base10_value);
          } break;
        case 1060: { //Primitive clear
            SerialFunctionPrimitiveClear(serial_base10_value);
          } break;
        case 1080: { //Primitive pulse
            //temporary wrong function
            uint16_t temp_array[22] = {1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
            //{65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535, 65535};
            for (uint16_t a = 0; a < serial_base10_value; a++) {
              PrintheadInkjetBurstW(temp_array);
            }
            //to be added
          } break;
        case 1100: { //shiftregister value
            PrintheadShiftSetW(serial_base10_value);
          } break;
        default: { //no valid code
            Serial.println("OP Error"); //print op error
            //perhaps make this more advanced than it is right now <---
          } break;
      } //end of execute command -------------------------------------------------


      //drain all used commands from the buffer, shift previous data to start of buffer and and fill open spots with 0
      int16_t temp_shift_position = temp_end_of_line;
      for (int16_t d = 0; d < serial_buffer_read_end; d++) {
        temp_shift_position ++; //end of line was on the last character of the line, move to the next
        if (temp_shift_position < serial_buffer_read_end) { //if end of line has not overtaken end of buffer, fill with character still in buffer
          serial_buffer[d] = serial_buffer[temp_shift_position];
        }
        else {
          serial_buffer[d] = 0;
        }
      }
      //Serial.print("old end: "); Serial.println(serial_buffer_read_end);
      serial_buffer_read_end -= temp_end_of_line; //subtract the number of characters shifted from end of the buffer
      serial_buffer_read_end --; //subtract 1 extra for value reasons
      //Serial.print("new end: "); Serial.println(serial_buffer_read_end);
    } //while(1) loop to read all lines in a buffer
  } //if serial was available
} //function end

//completely clears the buffer
void SerialClearBuffer() {

}



//takes an array of 3 chars and returns the op code value if there is one, returns -1 if no op code was found
int16_t SerialGetOpCode(char temp_input[]) {
  for (uint16_t p = 0; p < serial_op_codes; p++) { //walk through the list of op codes
    for (uint8_t c = 0; c < 4; c++) { //see if the value matches
      if (c == 3) { //if c == 3, there have been 3 characters matched and the op code was matched
        //Serial.print("OP code Match: "); Serial.println(serial_op_code_value[p]);
        return serial_op_code_value[p]; //return the op code
      }
      else {
        //see if the current character matches
        if (temp_input[c] != serial_op_code_name[p][c]) {
          break; //if it does not match, end this instance of testing and go to the next
        }
      }
    }

  }
  return -1;
}

//reads a character and determines if it is a valid op code character
uint8_t SerialIsCharacter(char temp_input) {
  if (temp_input >= 'A' && temp_input <= 'Z') return 1;
  return 0; //no match found
}

//reads a character and determines if it is a valid base 10 character
uint8_t SerialIsB10(char temp_input) {
  if (temp_input >= '0' && temp_input <= '9') return 1;
  if (temp_input == '-') return 1; //negative is accepted in base 10 mode
  return 0; //no match found
}

//reads a character and determines if it is a valid base 64 character
uint8_t SerialIsB64(char temp_input) {
  if (temp_input >= 'A' && temp_input <= 'Z') return 1;
  if (temp_input >= 'a' && temp_input <= 'z') return 1;
  if (temp_input >= '0' && temp_input <= '9') return 1;
  if (temp_input == '+') return 1;
  if (temp_input == '/') return 1;
  return 0; //no match found
}

//looks for a character that marks the end of a line. Returns 0 if not, and 1 if end character
uint8_t SerialLineEndCharacter(char temp_input) {
  if (temp_input == '\n') return 1; //new line character marks end
  else if (temp_input == '\r') return 1; //carriage return marks end
  else if (temp_input == '&') return 1; //ampersan marks end
  else return 0; //if none of the above are true, it is not an end character
}


int8_t SerialB10Lookup(char c) {
  if (c >= '0' && c <= '9') return c - 48; //0-9
  return -1; //-1
}

//Convert a base 64 character to decimal, (returns -1 if no valid character is found)
int8_t SerialB64Lookup(char c) {
  if (c >= 'A' && c <= 'Z') return c - 'A'; //0-25
  if (c >= 'a' && c <= 'z') return c - 71; //26-51
  if (c >= '0' && c <= '9') return c + 4; //52-61
  if (c == '+') return 62; //62
  if (c == '/') return 63; //63
  return -1; //-1
}
